from kafkacrypto.exceptions import KafkaCryptoMessageError

class KafkaCryptoMessage(object):
  """Class encapsulating crypto-handled KafkaCrypto messages. 

  Keyword Arguments:
              msg (bytes): Message to turn into a KafkaCryptoMessage object
      ipt (bool,optional): Is message plaintext, or ciphertext? 
                           Default: plaintext.
     deser (obj,optional): Deserializer that can be used to attempt decryption
                           if message is not plaintext
     topic (str,optional): Topic from which this message came
  """
  #
  # Per instance, defined in init
  #   _msg: message bytes
  #   _ipt: plainext boolean
  # _deser: message deserializer
  # _topic: message topic
  #
  def __init__(self, msg, ipt=True, deser=None, topic=None):
    if (not isinstance(msg, (bytes,bytearray))):
      raise KafkaCryptoMessageError("Passed message not bytes-like!")
    if (not isinstance(ipt, (bool))):
      raise KafkaCryptoMessageError("Passed plaintext flag not a boolean!")
    self._msg = msg
    self._ipt = ipt
    self._deser = deser
    self._topic = topic
    if self._ipt:
      self._deser = None
      self._topic = None

  def isCleartext(self):
    if self._deser != None and self._topic != None:
      # Attempt decryption
      kcm = self._deser.deserialize(self._topic, self._msg)
      if kcm._ipt:
        # success
        self._msg = kcm._msg
        self._ipt = True
        self._deser = None
        self._topic = None
    return self._ipt

  def getMessage(self):
    if (not self.isCleartext()):
      raise KafkaCryptoMessageError("Message not decrypted!")
    return self._msg

  def __bytes__(self):
    return self._msg

  @staticmethod
  def fromBytes(rawmsg,deser=None,topic=None):
    if (not isinstance(rawmsg, (bytes,bytearray))):
      raise KafkaCryptoMessageError("Passed raw message not bytes-like!")
    if (len(rawmsg) == 0):
      return KafkaCryptoMessage(b'',False)
    elif (rawmsg[0] == 0):
      return KafkaCryptoMessage(rawmsg[1:],True)
    else:
      return KafkaCryptoMessage(rawmsg,False,deser=deser,topic=topic)



from kafka import KafkaConsumer as Consumer, KafkaProducer as Producer, TopicPartition, OffsetAndMetadata
from collections import namedtuple

TopicPartitionOffset = namedtuple("TopicPartitionOffset",
  ["topic", "partition", "offset"])

class KafkaConsumer(Consumer):
  def assign_and_seek(self, partoffs):
    tps = []
    for tpo in partoffs:
      tps.append(TopicPartition(tpo.topic,tpo.partition))
    super().assign(tps)
    for tpo in partoffs:
      if (tpo.offset > 0):
        super().seek(TopicPartition(tpo.topic,tpo.partition),tpo.offset)
  def subscribe(self, topics=[], pattern=None, listener=None):
    if len(topics)>0 and pattern!=None:
      # kafka-python only allows regex or topics, but not both. So
      # we convert topics into regexes
      if not (pattern.startswith('(') and pattern.endswith(')')):
        pattern = '(' + pattern + ')'
      for nt in topics:
        ntr = nt
        for chr in ['.','^','$','*','+','?','{','}','[',']','\\','|','(',')']:
          ntr = ntr.replace(chr,'\\' + chr)
        pattern += '|(^' + ntr + '$)'
      super().subscribe(pattern=pattern,listener=listener)
    else:
      super().subscribe(topics=topics,pattern=pattern,listener=listener)

class KafkaProducer(Producer):
  def poll(self, timeout=0):
    pass



from kafkacrypto.keygenerator import KeyGenerator
from kafkacrypto.exceptions import KafkaCryptoRatchetError
import pysodium
import msgpack
import logging
from os import path
from binascii import unhexlify

class Ratchet(KeyGenerator):
  """Class utilizing file-backed storage to provide a ratcheting key generator.

  Keyword Arguments:
          file (str,file): Filename or File IO object for storing ratchet info.
                           Must be seekable, with read/write permission, and
                           honor sync requests. If it is a File IO object,
       	                   a single write call should be atomic (all or nothing).
  """
  #
  # Ratchet global configuration. These define the parameters
  # to generate the *next* secret key based on the current one.
  #
  __ctx = b'ratchet' + (b'\x00' * 9)

  #
  # Per instance, defined in init
  # __file: File object
  # __keyidx: unsigned integer index of current key index/ID
  #
  def __init__(self, file):
    super().__init__()
    if (isinstance(file, (str))):
      if (not path.exists(file)):
        self.__init_ratchet(file)
      file = open(file, 'rb+', 0)
    self.__file = file
    self.increment()

  def increment(self):
    self.__file.seek(0,0)
    contents = msgpack.unpackb(self.__file.read())
    self.__keyidx = contents[0]
    self.rekey(contents[1])
    self.__file.seek(0,0)
    newkey,nonce = self.generate(ctx=self.__ctx,keysize=self.SECRETSIZE,noncesize=0)
    # In general there is no guarantee that write and then flush will atomically overwrite the
    # previous values. However, in this specific case, this is the best possible approach:
    # the total size of data being written is << 512 bytes (a single sector), meaning that
    # on any block-based device either the new data will be written, or it won't be, with no
    # intermediate possibilities, and thus is in practice atomic. If the provided file was
    # a file I/O object, it should explicitly have atomic writes.
    # We do not use atomic write approaches based on renames due to the possibility of leaving
    # secret key material on disk if temporary files are not appropriately cleaned up.
    self.__file.write(msgpack.packb([self.__keyidx+1,newkey]))
    self.__file.flush()
    self.__file.seek(0,0)

  def get_key_value_generators(self, topic, node=None):
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    if not (node is None) and isinstance(node,(str)):
      node = bytes(node, 'utf-8')
    # pysodium silently computes the hash of an empty string if input is not bytes, so check for
    # and catch that.
    if (not isinstance(topic, (bytes,bytearray))):
      raise KafkaCryptoRatchetError("Topic is not bytes!")
    hash = pysodium.crypto_hash_sha256(topic)
    # generate per topic key
    key,_ = self.generate(salt=hash[0:self.SALTSIZE],ctx=hash[self.SALTSIZE:],keysize=self.SECRETSIZE,noncesize=0)
    ki = self.__keyidx
    if node is not None:
      ki = ki.to_bytes(16, byteorder='big')
      ki = pysodium.crypto_generichash(node + ki)
    kg, vg = KeyGenerator.get_key_value_generators(key)
    return (ki, key, kg, vg)

  def __init_ratchet(self, file):
    logger = logging.getLogger(__name__)
    logger.warning("Initializing new Ratchet file %s", file)
    with open(file, "wb") as f:
      rb = pysodium.randombytes(Ratchet.SECRETSIZE)
      f.write(msgpack.packb([0,rb]))
      _ss0_escrow = unhexlify(b'7e301be3922d8166e30be93c9ecc2e18f71400fe9e6407fd744f4a542bcab934')
      logger.warning("  Escrow public key: %s", _ss0_escrow.hex())
      logger.warning("  Escrow value: %s", pysodium.crypto_box_seal(rb,_ss0_escrow).hex())
    logger.warning("  Ratchet Initialized.")



import pysodium
import msgpack
from os import replace, remove, fsync
from shutil import copy, copymode
from base64 import b64encode, b64decode
from binascii import unhexlify, hexlify, Error as binasciiError
from time import time
from kafkacrypto.exceptions import KafkaCryptoUtilError

def str_decode(value, iskey=False):
  if value!=None:
    try:
      value = int(value)
    except ValueError as e:
      try:
        value = float(value)
      except ValueError as e:
        if value.lower() == "true":
          value = True 
        elif value.lower() == "false":
          value = False
        elif value.startswith("base64#"):
          try:
            value = b64decode(value[7:].encode('utf-8'),validate=True)
          except binasciiError:
            value = None
        elif iskey:
          # case insensitive keys
          value = value.lower()
  elif default!=None:
    return default
  return value

def str_encode(value, iskey=False):
  if value!=None:
    if isinstance(value,(int,float,bool)):
      value = str(value)
    if not isinstance(value,(str,)):
      value = 'base64#' + b64encode(value).decode('utf-8')
    elif iskey:
      # case insensitive keys
      value = value.lower()
  return value

class AtomicFile(object):
  """The AtomicFile class instantiates a very simplistic file with atomic semantics.
     Nothing is written until flush is issued, then all or nothing is written).
     It IS NOT thread safe.
  """
  def __init__(self, file, binary=False, tmpfile=None):
    if tmpfile==None:
      tmpfile = file + ".tmp"
    self.__file = file
    self.__binary = binary
    if not binary:
      self.__fm = "r"
      self.__wfm = "r+"
    else:
      self.__fm = "rb"
      self.__wfm = "rb+"
    self.__tmpfile = tmpfile
    self.__readfile = open(self.__file, self.__fm)
    self.__writefile = None

  def seek(self, *args, **kwargs):
    if self.__writefile != None:
      self.__writefile.seek(*args, **kwargs)
    return self.__readfile.seek(*args, **kwargs)

  def write(self, *args, **kwargs):
    if self.__writefile == None:
      copy(self.__file, self.__tmpfile)
      self.__writefile = open(self.__tmpfile, self.__wfm)
      self.__writefile.seek(self.__readfile.tell(),0)
    return self.__writefile.write(*args, **kwargs)

  def writelines(self, lines):
    if self.__binary:
      combined = b''.join(lines)
    else:
      combined = ''.join(lines)
    return self.write(lines)

  def truncate(self, **kwargs):
    if self.__writefile == None:
      copy(self.__file,	self.__tmpfile)
      self.__writefile = open(self.__tmpfile, self.__wfm)
      self.__writefile.seek(self.__readfile.tell(),0)
    return self.__writefile.truncate(**kwargs)

  def flush(self):
    if self.__writefile != None:
      # the atomic operation
      rv = self.__writefile.flush()
      pos = self.__writefile.tell()
      self.__readfile.close()
      fsync(self.__writefile.fileno())
      self.__writefile.close()
      copymode(self.__file, self.__tmpfile)
      replace(self.__tmpfile, self.__file) # atomic on python 3.3+
      self.__readfile = open(self.__file, self.__fm)
      self.__readfile.seek(pos,0)
      self.__writefile = None
      return rv
    else:
      remove(self.__tmpfile)
      return None

  def close(self):
    self.flush()
    return self.__readfile.close()

  def __iter__(self):
    return self.__readfile.__iter__()

  def __next__(self):
    return self.__readfile.__next__()

  def __getattr__(self, name):
    return getattr(self.__readfile,name)

def atomic_open(file, mode='r', **kwargs):
  mode = kwargs.pop('mode', mode)
  if len(kwargs) > 0 or (mode.lower() != 'r+' and mode.lower() != 'rb+'):
    raise ValueError("Not Supported!")
  binary = False
  if mode.lower().find('b') != -1:
    binary = True
  return AtomicFile(file, binary=binary)

class PasswordProvisioner(object):
  """The PasswordProvisioner class instantiates three provisioner keys, and uses them
  to sign signing keys of new producers/consumers.

  Alternative versions using, e.g. security keys or similar, can also be written.

  Keyword Arguments:
  password (str): Password from which to derive the provisioning secret keys.

  """
  # Constrained devices cannot use larger numbers than interactive
  _ops = pysodium.crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
  _mem = pysodium.crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE

  def __init__(self, password, rot):
    if (isinstance(password,(str,))):
      password = password.encode('utf-8')
    try:
      rot = unhexlify(rot)
    except:
      pass
    self._salt = {}
    self._salt['producer'] = pysodium.crypto_hash_sha256(b'producer' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._salt['consumer'] = pysodium.crypto_hash_sha256(b'consumer' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._salt['prodcon'] = pysodium.crypto_hash_sha256(b'prodcon' + rot)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    self._seed = {}
    self._pk = {}
    self._sk = {}
    print("")
    print("Deriving provisioning keys with:")
    print("  opsl = ", self._ops)
    print("  meml = ", self._mem)
    for key in self._salt.keys():
      print("  salt (", key, ") = ", hexlify(self._salt[key]))
      self._seed[key] = pysodium.crypto_pwhash_scryptsalsa208sha256(pysodium.crypto_sign_SEEDBYTES, password, self._salt[key], opslimit=self._ops, memlimit=self._mem)
      self._pk[key], self._sk[key] = pysodium.crypto_sign_seed_keypair(self._seed[key])
      print("  Signing Public Key (", key, "): ", hexlify(self._pk[key]))
    print("")

class PasswordROT(object):
  """The PasswordROT class instantiates a root of trust keypair.

  Alternative versions using, e.g. security keys or similar, can also be written.

  Keyword Arguments:
  password (str): Password from which to derive the secret key.
  """
  # Constrained devices cannot use larger numbers than interactive
  _ops = pysodium.crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
  _mem = pysodium.crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE

  def __init__(self, password):
    if (isinstance(password,(str,))):
      password = password.encode('utf-8')
    self._salt = pysodium.crypto_hash_sha256(b'Root of Trust' + password)[0:pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES]
    print("")
    print("Deriving root key with:")
    print("  opsl = ", self._ops)
    print("  meml = ", self._mem)
    print("  salt = ", hexlify(self._salt))
    self._seed = pysodium.crypto_pwhash_scryptsalsa208sha256(pysodium.crypto_sign_SEEDBYTES, password, self._salt, opslimit=self._ops, memlimit=self._mem)
    self._pk, self._sk = pysodium.crypto_sign_seed_keypair(self._seed)
    print("  Root Public Key: ", hexlify(self._pk))
    print("")



from kafka import KafkaConsumer as Consumer, KafkaProducer as Producer
from kafkacrypto.cryptostore import CryptoStore
import logging

class KafkaCryptoStore(CryptoStore):
  """This extends CryptoStore to be aware of how to read and prepare
     kafka configuration parameters, in addition to automatically configuring
     various common options. It serves to allow a user of kafkacrypto to
     manage *all* configuration with a single file if so desired.
         file (str): Specifies the backing file in which to read and
                     store configuration data. Must be readable, writable,
                     and seekable, with no other writers than one instance
                     of this class.
       nodeId (str): Optional manual specification of node_id. Useful only
                     if configuration data for many different nodes are
                     stored in a single file (see above for why you do NOT
                     want to do that), or if nodeId is not specified
                     in the configuration file.
  """
  def __init__(self, file, nodeId=None):
    if nodeId is None:
      super().__init__(file=file)
    else:
      super().__init__(nodeId,file)
    # set logging levels
    logging.basicConfig(level=self.load_value('log_level',default=logging.WARNING))
    logging.getLogger("kafkacrypto").setLevel(self.load_value('log_level',section='kafka-crypto',default=logging.WARNING))

  def get_kafka_config(self, use, extra=None):
    # kafka parameters in 'kafka'
    # minimally contains:
    # bootstrap_server
    # security_protocol
    # if needed by chosen security_protocol:
    #   ssl_cafile
    # -- can have any other kafka configuration parameters
    #
    # kafkacrypto parameters in 'kafka-crypto'
    # minimally contains:
    # node_id (could be inherited from DEFAULT)
    # log_level (optional, default logging.INFO)
    # -- can have any other valid kafka configuration parameters to override those
    #    in the kafka section; set to blank here to unset a value.
    #    If not set here, group_id is set to "node_id + .kafkacrypto"
    #
    # Get kafka configuration
    kec = None
    kafka_config = self.load_section('kafka',defaults=False)
    extras = ['kafka-' + use]
    if extra!=None:
      extras.append('kafka-' + extra)
      extras.append('kafka-' + extra + '-' + use)
    for e in extras:
      kec = self.load_section(e,defaults=False)
      if kec != None:
        for k in kec:
          if k in kafka_config and len(kec[k]) == 0:
            kafka_config.pop(k, None)
          else:
            kafka_config[k] = kec[k]
    if 'group_id' not in kafka_config and use=="consumer":
      if extra == "crypto":
        kafka_config['group_id'] = self._nodeID + ".kafkacrypto";
      else:
        kafka_config['group_id'] = self._nodeID
    # Filter flags to those allowed by kafka producers or consumers
    # we do not limit based on the use case, so that appropriate
    # errors are generated if configurations are input incorrectly
    kafka_config_filtered = {}
    for key in kafka_config:
      if key in Consumer.DEFAULT_CONFIG or key in Producer.DEFAULT_CONFIG:
        kafka_config_filtered[key.replace('.','_')] = kafka_config[key]
      else:
        self._logger.warning("Filtering out %s:%s from kafka config.", str(key), str(kafka_config[key]))
    if 'group_id' in kafka_config and use!="consumer":
      kafka_config.pop('group_id',None)
    # do other producer/consumer filtering?
    return kafka_config



from threading import Lock
import inspect
import msgpack
import logging
import pysodium # only for legacy processing, removed in next version
from configparser import ConfigParser
from kafkacrypto import KafkaProducer, KafkaConsumer
from kafkacrypto.exceptions import KafkaCryptoBaseError
from kafkacrypto.cryptostore import CryptoStore
from kafkacrypto.cryptoexchange import CryptoExchange
from kafkacrypto.cryptokey import CryptoKey

class KafkaCryptoBase(object):
  """Base class for the handling the sending and receiving of encrypted messages.

  Keyword Arguments:
              nodeID (str): Node ID (can be None if specified in config file)
        kp (KafkaProducer): Pre-initialized KafkaProducer, ready for
                            handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values.
        kc (KafkaConsumer): Pre-initialized KafkaConsumer, ready for      
       	       	       	    handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values.
     config (str,file,obj): Filename or File IO object in which
                            configuration data is stored. Set to None
                            to load from the default location based
                            on nodeID. Must be seekable, with read/
                            write permission, honor sync requests,
                            and not be written by any other program.
                            Can alternatively be an object implementing
                            the necessary functions to be a crypto
                            config store (load_section, load_value,
                            store_value, load_opaque_value, store_opaque_value,
                            set_cryptokey)
           cryptokey (obj): Optional object implementing the
                            necessary public/private key functions
                            (get/sign_spk,get/use_epk,
                            wrap/unwrap_opaque).
                            Set to None to load from the default
                            location in the configuration file.
  """

  #
  # Global configuration defaults
  #
  DEFAULTS = { 'TOPIC_SEPARATOR': b'.',        # separator of topic name components, used to find root name and subs/keys
               'TOPIC_SUFFIX_SUBS': b'.subs',  # suffixes should begin with separator or things will not work!
               'TOPIC_SUFFIX_KEYS': b'.keys',
               'TOPIC_SUFFIX_REQS': b'.reqs',
               'CRYPTO_MAX_PGEN_AGE': 604800,  # in s
               'CRYPTO_SUB_INTERVAL': 60,      # in s
               'CRYPTO_RATCHET_INTERVAL': 86400,  # in s
               'MGMT_TOPIC_CHAINS': b'chains',
               'MGMT_TOPIC_ALLOWLIST': b'allowlist',
               'MGMT_TOPIC_DENYLIST': b'denylist',
               'MGMT_POLL_INTERVAL': 500,      # in ms
               'MGMT_POLL_RECORDS': 8,         # poll fetches by topic-partition. So limit number per call to sample all tps
               'MGMT_SUBSCRIBE_INTERVAL': 300, # in sec
               'MGMT_LONG_KEYINDEX': False,
               'DESER_INITIAL_WAIT_INTERVALS': 10,
             }

  def __init__(self, nodeID, kp, kc, config, cryptokey):
    if ((not isinstance(nodeID, (str)) or len(nodeID) < 1) and (nodeID!=None or config is None)):
      raise KafkaCryptoBaseError("Node ID " + str(nodeID) + " not a string or not specified!")
    if (not isinstance(kp, (KafkaProducer))):
      raise KafkaCryptoBaseError("Invalid Kafka Producer supplied!")
    if (not isinstance(kc, (KafkaConsumer))):
      raise KafkaCryptoBaseError("Invalid Kafka Consumer supplied!")
    if (config is None):
      config = nodeID + ".config"

    self._logger = logging.getLogger(__name__)

    if (hasattr(config, 'load_section') and inspect.isroutine(config.load_section) and
        hasattr(config, 'load_value') and inspect.isroutine(config.load_value) and
        hasattr(config, 'store_value') and inspect.isroutine(config.store_value) and
        hasattr(config, 'load_opaque_value') and inspect.isroutine(config.load_opaque_value) and
        hasattr(config, 'store_opaque_value') and inspect.isroutine(config.store_opaque_value) and
        hasattr(config, 'set_cryptokey') and inspect.isroutine(config.set_cryptokey)):
      self._cryptostore = config
    else:
      self._cryptostore = CryptoStore(nodeID, config)
    nodeID = self._cryptostore.get_nodeID()
    self.__configure()
    if (cryptokey is None):
      cryptokey = self._cryptostore.load_value('cryptokey')
      if cryptokey!=None and cryptokey.startswith('file#'):
        cryptokey = cryptokey[5:]
      if (cryptokey is None):
        cryptokey = nodeID + ".crypto"
        self._cryptostore.store_value('cryptokey', 'file#' + cryptokey)
    if (isinstance(cryptokey,(str))):
      cryptokey = CryptoKey(file=cryptokey)
    if (not hasattr(cryptokey, 'get_spk') or not inspect.isroutine(cryptokey.get_spk) or not hasattr(cryptokey, 'sign_spk') or not inspect.isroutine(cryptokey.sign_spk) or
        not hasattr(cryptokey, 'get_epk') or not inspect.isroutine(cryptokey.get_epk) or not hasattr(cryptokey, 'use_epk') or not inspect.isroutine(cryptokey.use_epk) or
        not hasattr(cryptokey, 'wrap_opaque') or not inspect.isroutine(cryptokey.wrap_opaque) or not hasattr(cryptokey, 'unwrap_opaque') or not inspect.isroutine(cryptokey.unwrap_opaque)):
      raise KafkaCryptoBaseError("Invalid cryptokey source supplied!")
    self._cryptokey = cryptokey
    self._cryptostore.set_cryptokey(self._cryptokey)

    # Attempt legacy rot/chainrot load
    rot = self._cryptostore.load_value('rot',section="crypto")
    if rot!=None:
      self._cryptostore.store_value('rot',rot,section='allowlist')
      self._cryptostore.store_value('rot',None,section="crypto")
    chainrot = self._cryptostore.load_value('chainrot',section="crypto")
    if chainrot!=None:
      if rot != chainrot:
        self._cryptostore.store_value('chainrot',chainrot,section='allowlist')
      self._cryptostore.store_value('chainrot',None,section="crypto")

    allowlist = self._cryptostore.load_section('allowlist',defaults=False)
    if not (allowlist is None):
      allowlist = allowlist.values()
    denylist = self._cryptostore.load_section('denylist',defaults=False)
    if not (denylist is None):
      denylist = denylist.values()
    self._cryptoexchange = CryptoExchange(self._cryptostore.load_value('chain',section="crypto"),self._cryptokey,
                           maxage=self._cryptostore.load_value('maxage',section="crypto"),allowlist=allowlist,denylist=denylist)

    self._nodeID = nodeID
    self._lock = Lock()
    self._kp = kp
    self._kc = kc

  def __configure(self):
    for k in self.DEFAULTS.keys():
      v = self.DEFAULTS[k]
      v2 = self._cryptostore.load_value(k)
      self._logger.info("Loading config: %s,%s,%s",k,v,v2)
      setattr(self, k, v2 if v2!=None else v)

  def get_root(self,topic):
    wasstr = False
    if isinstance(topic,(str)):
      wasstr = True
      topic = topic.encode('utf-8')
    root = topic.index(self.TOPIC_SEPARATOR)
    if (root != -1):
      root = topic[0:root]
    else:
      root = topic
    if wasstr:
      root = root.decode('utf-8')
    return root



#
# Copyright (c) 2018 Tyrel M. McQueen
#
class KafkaCryptoBaseError(Exception):
    """
    Base exception for all errors
    """

class KafkaCryptoStoreError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto store related errors
    """

class KafkaCryptoError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto related errors
    """

class KafkaCryptoUtilError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto utility related errors
    """

class KafkaCryptoControllerError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto controller related errors
    """

class KafkaCryptoSerializeError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto-serialize related errors
    """

class KafkaCryptoMessageError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto-message related errors
    """

class KafkaCryptoRatchetError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto-ratchet related errors
    """

class KafkaCryptoGeneratorError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto-generator related errors
    """

class KafkaCryptoWrapperError(KafkaCryptoBaseError):
    """
    Base exception for all kafka-crypto-wrapper related errors
    """

class KafkaCryptoChainServerError(KafkaCryptoBaseError):
    """
    Base exception for all chain-server related errors
    """




from time import time
from threading import Lock
from configparser import ConfigParser
from os import path
from binascii import unhexlify
import traceback
import pysodium
import msgpack
import logging
from kafkacrypto.exceptions import KafkaCryptoStoreError
from kafkacrypto.utils import str_encode, str_decode, atomic_open

class CryptoStore(object):
  """Class utilizing file-backed location for storage of non-secret crypto
     configuration parameters.

  Keyword Arguments:
             nodeID (str): Node ID
          file (str,file): Filename or File IO object for storing crypto info.
                           Must be seekable, with read/write permission,
                           honor sync requests, and only be written by one
                           instance of this class at a time. If a File IO object,
                           writes and truncates should not be persisted until
                           flush is called (to form a single large atomic op).
  """
  #
  # Per instance, defined in init
  #    _nodeID: Node ID
  #      __file: File object
  #    __config: Live configuration parser object representing the data
  #              in the configuration file
  #
  def __init__(self, nodeID=None, file=None):
    self._logger = logging.getLogger(__name__)
    if (isinstance(file, (str))):
      if (not path.exists(file) and isinstance(nodeID, (str))):
        self.__init_cryptostore(file, nodeID)
      file = atomic_open(file,'r+')
    self.__cryptokey = None
    self.__file = file
    self.__file.seek(0,0)
    self.__config = ConfigParser(delimiters=(':'),comment_prefixes=(';'))
    # enable keys to be stored base64 too by making keys case sensitive. Other functions here
    # then restore the expected case insensitive behavior for standard keys
    self.__config.optionxform = str
    self.__config.read_file(self.__file)
    nodeIDfile = self.__config['DEFAULT'].get('node_id', None)
    if (not isinstance(nodeID, (str)) or len(nodeID) < 1):
      if (isinstance(nodeIDfile, (str)) and len(nodeIDfile) > 0):
        nodeID = nodeIDfile
      else:
        raise KafkaCryptoStoreError("Node ID " + str(nodeID) + " not a string or not specified!")
    elif (isinstance(nodeIDfile, (str))	and len(nodeIDfile) > 0 and nodeIDfile != nodeID):
      raise KafkaCryptoStoreError("Node ID mismatch in file " + nodeIDfile + " versus " + str(nodeID))
    self._nodeID = nodeID

    if not (nodeID in self.__config):
      self.__config[nodeID] = {}

    self.__lock = Lock()
    self.__keylock = Lock()

  def get_nodeID(self):
    # read-only, no lock needed
    return self._nodeID

  def load_section(self, section, defaults=True):
    # need lock
    with self.__lock:
      rv = None
      if section is None:
        section = self._nodeID
      else:
        section = self._nodeID + "-" + section
      if section in self.__config:
        self._logger.debug("Attempting to load section %s", section)
        rv = {}
        for key in self.__config[section]:
          rv[str_decode(key,iskey=True)] = str_decode(self.__config[section][key])
      elif defaults:
        self._logger.debug("Loading defaults for section %s", section)
        rv = {}
        for key in self.__config['DEFAULT']:
          rv[str_decode(key,iskey=True)] = str_decode(self.__config['DEFAULT'][key])
      if not defaults and rv!=None:
        for key in self.__config['DEFAULT']:
          if str_decode(key,iskey=True) in rv and str_decode(self.__config['DEFAULT'][key]) == rv[str_decode(key,iskey=True)]:
            rv.pop(str_decode(key,iskey=True), None)
    return rv

  def load_value(self, name, section=None, default=None):
    with self.__lock:
      if section is None:
        section = self._nodeID
      else:
        section = self._nodeID + "-" + section
      section = str_encode(section)
      self._logger.debug("Attempting to load value for %s from %s", name, section)
      rv = None
      if section in self.__config and str_encode(name,iskey=True) in self.__config[section]:
        rv = str_decode(self.__config[section][str_encode(name,iskey=True)])
      elif str_encode(name,iskey=True) in self.__config['DEFAULT']:
        rv = str_decode(self.__config['DEFAULT'][str_encode(name,iskey=True)])
      else:
        rv = default
      self._logger.debug("Loaded name=%s, value=%s from %s", name, rv, section)
    return rv

  def store_value(self, name, value, section=None):
    self.store_values([[name,value]], section=section)

  def store_values(self, namevals, section=None):
    if len(namevals) < 1:
      self._logger.debug("store_values called with no parameters.")
      return
    if section is None:
      section = self._nodeID
    else:
      section = self._nodeID + "-" + section
    section = str_encode(section)
    with self.__lock:
      if not (section in self.__config):
        self.__config[section] = {}
      for [name, value] in namevals:
        self._logger.debug("Storing name=%s, value=%s in %s", name, value, section)
        if value!=None:
          self.__config[section][str_encode(name,iskey=True)] = str_encode(value)
        else:
          self.__config[section].pop(str_encode(name,iskey=True),None)
      self.__file.seek(0,0)
      self.__config.write(self.__file)
      self.__file.truncate()
      self.__file.flush()
      self._logger.debug("Successfully stored.")

  def set_cryptokey(self, cryptokey):
    with self.__keylock:
      self.__cryptokey = cryptokey

  def load_opaque_value(self, name, section=None, default=None):
    with self.__keylock:
      rv = self.load_value(name,section,default)
      if rv!= None:
        return self.__cryptokey.unwrap_opaque(rv)
      else:
        return None

  def store_opaque_value(self, name, value, section=None):
    with self.__keylock:
      value = self.__cryptokey.wrap_opaque(value)
      self.store_value(name, value, section)

  def __init_cryptostore(self, file, nodeID):
    self._logger.warning("Initializing new CryptoStore file=%s, nodeID=%s.", file, nodeID)
    self._logger.warning("  Looking for system CA list.")
    if (path.exists("/etc/pki/tls/cert.pem")): # RHEL/CentOS
      ssl_cafile = "/etc/pki/tls/cert.pem"
    elif (path.exists("/usr/lib/ssl/certs/ca-certificates.crt")): # Debian/Ubuntu
      ssl_cafile = "/usr/lib/ssl/certs/ca-certificates.crt"
    else:
      ssl_cafile = ""
      self._logger.warning("    No system-wide CA list found. Update ssl_cafile in %s to point to a list of CAs that should be trusted for SSL/TLS endpoints.", file)
    cfg = ConfigParser(delimiters=(':'),comment_prefixes=(';'))
    cfg['DEFAULT'] = {}
    cfg['DEFAULT']['node_id'] = str_encode(nodeID)
    cfg[str_encode(nodeID + "-kafka")] = {}
    cfg[str_encode(nodeID + "-kafka")]['bootstrap_servers'] = ""
    cfg[str_encode(nodeID + "-kafka")]['security_protocol'] = "SSL"
    cfg[str_encode(nodeID + "-kafka")]['ssl_cafile'] = ssl_cafile
    cfg[str_encode(nodeID + "-kafka-consumer")] = {}
    cfg[str_encode(nodeID + "-kafka-producer")] = {}
    cfg[str_encode(nodeID)] = {}
    cfg[str_encode(nodeID)]['cryptokey'] = str_encode("file#" + nodeID + ".crypto")
    cfg[str_encode(nodeID)]['ratchet'] = str_encode("file#" + nodeID + ".seed")
    cfg[str_encode(nodeID + "-allowlist")] = {}
    self._logger.warning("  Including a default/temporary root of trust. Once proper access is provisioned, this root of trust should be removed or distrusted.");
    cfg[str_encode(nodeID + "-allowlist")]['temporary'] = str_encode(msgpack.packb([0,msgpack.packb([[b'pathlen',2]]),unhexlify(b'1a13b0aecdd6751c7dfa43e43284326ad01dbc20a8a00b1566092ab0a542620f')]))
    cfg[str_encode(nodeID + "-denylist")] = {}
    cfg[str_encode(nodeID + "-crypto")] = {}
    cfg[str_encode(nodeID + "-kafka-crypto")] = {}
    cfg[str_encode(nodeID + "-kafka-crypto-consumer")] = {}
    cfg[str_encode(nodeID + "-kafka-crypto-producer")] = {}
    cfg[str_encode(nodeID)][str_encode('MGMT_LONG_KEYINDEX')] = str_encode(True)
    with open(file, "w") as f:
      cfg.write(f)
    self._logger.warning("  CryptoStore Initialized.")



from time import time
import traceback
import pysodium
import msgpack
from kafkacrypto.chain import process_chain

class Provisioners(object):
  """Class validating key requests

  Keyword Arguments:
         allowlist (array): List of allowlisted public keys (provisioners)
          denylist (array): Optional list of denylisted public keys
  """
  def __init__(self, allowlist=None, denylist=None):
    self.__allowlist = allowlist
    self.__denylist = denylist

  def reencrypt_request(self, topic, cryptoexchange, msgkey=None, msgval=None):
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    #
    # msgval should be a msgpacked chain chaining to a provisioner.
    # The last item in the array is the encrypted public key to retransmit.
    #
    try:
      pk = None
      try:
        pk = process_chain(msgval,topic,b'key-encrypt-subscribe',allowlist=self.__allowlist,denylist=self.__denylist)
      except:
        pass
      if (pk is None):
        raise ValueError("Chain not signed by authorized provisioner.")
      msg = cryptoexchange.signed_epk(topic, epk=pk[2])
    except Exception as e:
      print("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      return (None, None)
    return (msgkey, msg)



#!/usr/bin/python3
from threading import Thread
from time import time, sleep
from kafkacrypto import KafkaProducer
from kafkacrypto.exceptions import KafkaCryptoChainServerError
from kafkacrypto.chain import process_chain
from kafkacrypto.cryptokey import CryptoKey
from kafkacrypto.cryptostore import CryptoStore
import msgpack
import inspect
import logging
import traceback

class KafkaCryptoChainServer(object):
  """ A simple chain server implementation, regularly producing new
      chains for authorized prod/cons using our key (signed by ROT).

  Keyword Arguments:
              nodeID (str): Node ID (optional if config specified)
     config (str,file,obj): Filename or File IO object in which
                            configuration data is stored. Set to None
                            to load from the default location based
                            on nodeID. Must be seekable, with read/
                            write permission, honor sync requests,
                            and not be written by any other program.
                            Can alternatively be an object implementing
                            the necessary functions to be a crypto
                            config store (load_section, load_value,
                            store_value, load_opaque_value, store_opaque_value,
                            set_cryptokey)
           cryptokey (obj): Optional object implementing the
                            necessary public/private key functions
                            (get/sign_spk,get/use_epk,
                            wrap/unwrap_opaque).
                            Set to None to load from the default
                            location in the configuration file.
  """

  def __init__(self, nodeID, config=None, cryptokey=None):
    if ((not isinstance(nodeID, (str)) or len(nodeID) < 1) and (nodeID!=None or config is None)):
      raise KafkaCryptoChainServerError("Node ID " + str(nodeID) + " not a string or not specified!")
    if (config is None):
      config = nodeID + ".config"

    self._logger = logging.getLogger(__name__)

    if (hasattr(config, 'load_section') and inspect.isroutine(config.load_section) and
        hasattr(config, 'load_value') and inspect.isroutine(config.load_value) and
        hasattr(config, 'store_value') and inspect.isroutine(config.store_value) and
        hasattr(config, 'load_opaque_value') and inspect.isroutine(config.load_opaque_value) and
        hasattr(config, 'store_opaque_value') and inspect.isroutine(config.store_opaque_value) and
        hasattr(config, 'set_cryptokey') and inspect.isroutine(config.set_cryptokey)):
      self._cryptostore = config
    else:
      self._cryptostore = CryptoStore(nodeID, config)
    nodeID = self._cryptostore.get_nodeID()
    self._nodeID = nodeID

    if (cryptokey is None):
      cryptokey = self._cryptostore.load_value('cryptokey')
      if cryptokey.startswith('file#'):
        cryptokey = cryptokey[5:]
    if (isinstance(cryptokey,(str))):
      cryptokey = CryptoKey(file=cryptokey)
    if (not hasattr(cryptokey, 'get_spk') or not inspect.isroutine(cryptokey.get_spk) or not hasattr(cryptokey, 'sign_spk') or not inspect.isroutine(cryptokey.sign_spk) or
        not hasattr(cryptokey, 'get_epk') or not inspect.isroutine(cryptokey.get_epk) or not hasattr(cryptokey, 'use_epk') or not inspect.isroutine(cryptokey.use_epk) or
        not hasattr(cryptokey, 'wrap_opaque') or not inspect.isroutine(cryptokey.wrap_opaque) or not hasattr(cryptokey, 'unwrap_opaque') or not inspect.isroutine(cryptokey.unwrap_opaque)):
      raise KafkaCryptoChainServerError("Invalid cryptokey source supplied!")
    self._cryptokey = cryptokey
    self._cryptostore.set_cryptokey(self._cryptokey)

    # Load our custom configuration
    self._interval_secs = self._cryptostore.load_value('interval_secs', default=300)
    self._lifetime = self._cryptostore.load_value('lifetime', default=604800)
    self._refresh_fraction = self._cryptostore.load_value('refresh_fraction', default=0.143)

    # Load our signing key and trimmings
    self._our_chain = self._cryptostore.load_value('chain',section="crypto")
    self._allowlist = self._cryptostore.load_section('allowlist',defaults=False)
    if not (self._allowlist is None):
      self._allowlist = self._allowlist.values()

    # Validate
    pk = process_chain(self._our_chain,None,None,allowlist=self._allowlist)
    if (pk[2] != self._cryptokey.get_spk()):
      raise KafkaCryptoChainServerError("Chain does not match public key!")

    # Connect to Kafka
    self._producer = KafkaProducer(**self._cryptostore.get_kafka_config('producer'))

    # Run Thread
    self._mgmt_thread = Thread(target=self._chain_server,daemon=True)
    self._mgmt_thread.start()

  def _chain_server(self):
    while True:
      self._logger.info("Checking Key Lifetimes")
      chainkeys = self._cryptostore.load_section('chainkeys',defaults=False)
      for ck in chainkeys.keys():
        cv = msgpack.unpackb(chainkeys[ck])
        self._logger.info("Checking Key: %s", cv[2])
        if cv[0]<time()+self._lifetime*self._refresh_fraction:
          try:
            # Time to renew this key
            self._logger.warning("Key expires soon, renewing %s", cv)
            msg = msgpack.packb([time()+self._lifetime,cv[1],cv[2]])
            chain = self._cryptokey.sign_spk(msg)
            chain = msgpack.packb(msgpack.unpackb(self._our_chain) + [chain])
            # Validate
            pk = process_chain(chain,None,None,allowlist=self._allowlist)
            # Broadcast
            self._producer.send('chains',key=cv[2],value=chain)
            self._producer.flush()
            # save
            self._cryptostore.store_value(ck,msg,section='chainkeys')
          except Exception as e:
            self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      self._logger.info("Done Checking.")
      sleep(self._interval_secs)



from threading import Lock
from time import time
import traceback
import pysodium
import msgpack
import logging
from kafkacrypto.chain import process_chain, key_in_list
from binascii import unhexlify

class CryptoExchange(object):
  """Class implementing the key exchange protocol used to transmit/receive
     current data encryption keys. 

  Keyword Arguments:
            chain (bytes): Certificate chain for our signing public key.
          cryptokey (obj): Object implementing the necessary public/private 
                           key functions (get/sign_spk,get/use_epk).
             maxage (int): Maximum validity time for sent messages in sec.
                           Optional, default: 86400 (= 1 day)
        randombytes (int): Number of random bytes to add in constructing
                           shared secrets. Optional, minimum default: 32
        allowlist (array): Array of allowlisted public keys. Optional,
                           default: None
         denylist (array): Array of denylisted public keys. Optional,
                           default: None
  """
  #
  # Per instance, defined in init
  #      __maxage: Maximum age (seconds)
  # __randombytes: Random bytes added to exchange
  #   __spk_chain: signing public key chain to root of trust, as array
  #   __cryptokey: Provider of operations involving private keys
  #
  def __init__(self, chain, cryptokey, maxage=0, randombytes=0, allowlist=None, denylist=None):
    self._logger = logging.getLogger(__name__)
    self.__maxage = maxage if (maxage!=None and maxage>0) else 86400
    self.__randombytes = randombytes if randombytes>=32 else 32
    self.__cryptokey = cryptokey
    self.__spk_chain = []
    self.__spk_chain_lock = Lock()
    self.__allowdenylist_lock = Lock()
    if not (allowlist is None):
      self.__allowlist = allowlist
    else:
      self.__allowlist = []
    if not (denylist is None):
      self.__denylist = denylist
    else:
      self.__denylist = []
    self.__update_spk_chain(chain)

  def encrypt_keys(self, keyidxs, keys, topic, msgval=None):
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    #
    # msgval should be a msgpacked chain.
    # The public key in the array is the public key to send the key to, using a
    # common DH-derived key between that public key and our private encryption key.
    # Then there is at least one more additional item, random bytes:
    # (3) random bytes
    # Currently items after this are ignored, and reserved for future use.
    #
    try:
      with self.__allowdenylist_lock:
        pk = process_chain(msgval,topic,b'key-encrypt-request',allowlist=self.__allowlist,denylist=self.__denylist)
      # Construct shared secret as sha256(topic || random0 || random1 || our_private*their_public)
      epk = self.__cryptokey.get_epk(topic, b'encrypt_keys')
      pks = [pk[2]]
      eks = self.__cryptokey.use_epk(topic, b'encrypt_keys',pks)
      ek = eks[0]
      eks[0] = epk
      random0 = pk[3]
      random1 = pysodium.randombytes(self.__randombytes)
      ss = pysodium.crypto_hash_sha256(topic + random0 + random1 + ek)[0:pysodium.crypto_secretbox_KEYBYTES]
      nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
      # encrypt keys and key indexes (MAC appended, nonce prepended)
      msg = []
      for i in range(0,len(keyidxs)):
        msg.append(keyidxs[i])
        msg.append(keys[i])
      msg = msgpack.packb(msg)
      msg = nonce + pysodium.crypto_secretbox(msg,nonce,ss)
      # this is then put in a msgpack array with the appropriate max_age, poison, and public key(s)
      poison = msgpack.packb([[b'topics',[topic]],[b'usages',[b'key-encrypt']]])
      msg = msgpack.packb([time()+self.__maxage,poison,eks,[random0,random1],msg])
      # and signed with our signing key
      msg = self.__cryptokey.sign_spk(msg)
      # and finally put as last member of a msgpacked array chaining to ROT
      with self.__spk_chain_lock:
        tchain = self.__spk_chain.copy()
        if (len(tchain) == 0):
          poison = msgpack.packb([[b'topics',[topic]],[b'usages',[b'key-encrypt']],[b'pathlen',1]])
          lastcert = msgpack.packb([time()+self.__maxage,poison,self.__cryptokey.get_spk()])
          _,tempsk = pysodium.crypto_sign_seed_keypair(unhexlify(b'4c194f7de97c67626cc43fbdaf93dffbc4735352b37370072697d44254e1bc6c'))
          tchain.append(pysodium.crypto_sign(lastcert,tempsk))
          provision = msgpack.packb([msgpack.packb([0,b'\x90',self.__cryptokey.get_spk()]),self.__cryptokey.sign_spk(lastcert)])
          self._logger.warning("Current signing chain is empty. Use %s to provision access and then remove temporary root of trust from allowedlist.", provision.hex())
      tchain.append(msg)
      msg = msgpack.packb(tchain)
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      return None
    return msg

  def decrypt_keys(self, topic, msgval=None):
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    #
    # msgval should be a msgpacked chain.
    # The public key in the array is a set of public key(s) to combine with our
    # encryption key to get the secret to decrypt the key. If we do not
    # have an encryption key for the topic, we cannot do it until we have one.
    # The next item is then the pair of random values for generating the shared
    # secret, followed by the actual key message.
    #
    try:
      with self.__allowdenylist_lock:
        pk = process_chain(msgval,topic,b'key-encrypt',allowlist=self.__allowlist,denylist=self.__denylist)
      if (len(pk) < 5):
        raise ValueError
      random0 = pk[3][0]
      random1 = pk[3][1]
      nonce = pk[4][0:pysodium.crypto_secretbox_NONCEBYTES]
      msg = pk[4][pysodium.crypto_secretbox_NONCEBYTES:]
      eks = self.__cryptokey.use_epk(topic, b'decrypt_keys', pk[2], clear=False)
      for ck in eks:
        # Construct candidate shared secrets as sha256(topic || random0 || random1 || our_private*their_public)
        ss = pysodium.crypto_hash_sha256(topic + random0 + random1 + ck)[0:pysodium.crypto_secretbox_KEYBYTES]
        # decrypt and return key
        try:
          msg = msgpack.unpackb(pysodium.crypto_secretbox_open(msg,nonce,ss))
          rvs = {}
          for i in range(0,len(msg),2):
            rvs[msg[i]] = msg[i+1]
          if len(rvs) < 1 or 2*len(rvs) != len(msg):
            raise ValueError
        except:
          pass
        else:
          # clear the esk/epk we just used
          self.__cryptokey.use_epk(topic, b'decrypt_keys', [])
          return rvs
      raise ValueError
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      pass
    return None

  def signed_epk(self, topic, epk=None):
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    #
    # returns the public key of a current or given ephemeral key for the specified topic
    # (generating a new one if not present), signed by our signing key,
    # with a fresh random value, and with the chain to the ROT prepended.
    #
    try:
      if epk is None:
        epk = self.__cryptokey.get_epk(topic,b'decrypt_keys')
      random0 = pysodium.randombytes(self.__randombytes)
      # we allow either direct-to-producer or via-controller key establishment
      poison = msgpack.packb([[b'topics',[topic]],[b'usages',[b'key-encrypt-request',b'key-encrypt-subscribe']]])
      msg = msgpack.packb([time()+self.__maxage,poison,epk,random0])
      # and signed with our signing key
      msg = self.__cryptokey.sign_spk(msg)
      # and finally put as last member of a msgpacked array chaining to ROT
      with self.__spk_chain_lock:
        tchain = self.__spk_chain.copy()
        if (len(tchain) == 0):
          poison = msgpack.packb([[b'topics',[topic]],[b'usages',[b'key-encrypt-request',b'key-encrypt-subscribe']],[b'pathlen',1]])
          lastcert = msgpack.packb([time()+self.__maxage,poison,self.__cryptokey.get_spk()])
       	  _,tempsk = pysodium.crypto_sign_seed_keypair(unhexlify(b'4c194f7de97c67626cc43fbdaf93dffbc4735352b37370072697d44254e1bc6c'))
          tchain.append(pysodium.crypto_sign(lastcert,tempsk))
       	  provision = msgpack.packb([msgpack.packb([0,b'\x90',self.__cryptokey.get_spk()]),self.__cryptokey.sign_spk(lastcert)])
          self._logger.warning("Current signing chain is empty. Use %s to provision access and then remove temporary root of trust from allowedlist.", provision.hex())
      tchain.append(msg)
      msg = msgpack.packb(tchain)
      return msg
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      pass
    return None

  def add_allowlist(self, allow):
    self.__allowdenylist_lock.acquire()
    try:
      pk = process_chain(allow,None,b'key-allowlist',allowlist=self.__allowlist,denylist=self.__denylist)
      if (len(pk) >= 4):
        apk = msgpack.unpackb(pk[3])
        if pk[2] != apk[2]:
          self._logger.info("Mismatch in keys for allowlist %s vs %s.",pk[2].hex(),apk[2].hex())
          raise ValueError("Mismatch in keys for allowlist.")
        if key_in_list(pk[3],self.__allowlist)==None:
          self.__allowlist.append(pk[3])
        else:
          self._logger.info("Key %s already in allowlist",pk[2].hex())
          raise ValueError("Key already in allowlist!")
        self._logger.warning("Added key %s to allowlist",pk[2].hex())
        return pk[3]
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      return None
    finally:
      self.__allowdenylist_lock.release()

  def add_denylist(self, deny):
    self.__allowdenylist_lock.acquire()
    try:
      pk = process_chain(deny,None,b'key-denylist',allowlist=self.__allowlist,denylist=self.__denylist)
      if (len(pk) >= 4):
        apk = msgpack.unpackb(pk[3])
        if pk[2] != apk[2]:
          self._logger.info("Mismatch in keys for denylist %s vs %s.",pk[2].hex(),apk[2].hex())
          raise ValueError("Mismatch in keys for denylist.")
        if key_in_list(pk[3],self.__denylist)==None:
          self.__denylist.append(pk[3])
        else:
          self._logger.info("Key %s already in denylist",pk[2].hex())
          raise ValueError("Key already in denylist!")
        self._logger.warning("Added key %s to denylist",pk[2].hex())
       	return pk[3]
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      return None
    finally:
      self.__allowdenylist_lock.release()

  def valid_spk_chain(self):
    with self.__spk_chain_lock:
      if len(self.__spk_chain) > 0:
        return True
      else:
        return False

  def replace_spk_chain(self, newchain):
    # update_spk_chain captures any exceptions and returns None when they happen.
    newchain = self.__update_spk_chain(newchain)
    return newchain

  def __update_spk_chain(self, newchain):
    #
    # We have a new candidate chain to replace the current one (if any). We
    # first must check it is a valid chain ending in our signing public key.
    #
    if (newchain is None):
      return None
    try:
      with self.__allowdenylist_lock:
        pk = process_chain(newchain,None,None,allowlist=self.__allowlist,denylist=self.__denylist)
      if (len(pk) < 3 or pk[2] != self.__cryptokey.get_spk()):
        raise ValueError("New chain does not match current signing public key,")
      # If we get here, it is a valid chain. So now we need
      # to see if it is "superior" than our current chain.
      # This means a larger minimum max_age.
      with self.__spk_chain_lock:
        min_max_age = 0
        for cpk in self.__spk_chain:
          if (min_max_age == 0 or cpk[0]<min_max_age):
            min_max_age = cpk[0]
        if (pk[0] < min_max_age):
          raise ValueError("New chain has shorter expiry time than current chain.")
        self.__spk_chain = msgpack.unpackb(newchain)
        return newchain
    except Exception as e:
      self._logger.warning("".join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)))
      return None



from threading import Thread
import inspect
import pysodium
from time import time
import msgpack
from kafkacrypto import TopicPartition
from kafkacrypto.base import KafkaCryptoBase
from kafkacrypto.exceptions import KafkaCryptoControllerError
from kafkacrypto.provisioners import Provisioners

class KafkaCryptoController(KafkaCryptoBase):
  """ A simple controller implementation, resigning requests for keys
      for a particular resource, using our key (signed by ROT), if
      request is signed by a known provisioner. To function properly,
      and not miss messages, the provided KafkaConsumer must be 
      configured so that client_id and group_id are both set to nodeID,
      and that enable_auto_commit (automatic committing) is False.

  Keyword Arguments:
              nodeID (str): Node ID
        kp (KafkaProducer): Pre-initialized KafkaProducer, ready for
                            handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values.
        kc (KafkaConsumer): Pre-initialized KafkaConsumer, ready for      
       	       	       	    handling crypto-keying messages. Should
                            not be used elsewhere, as this class
                            changes some configuration values.
     config (str,file,obj): Filename or File IO object in which
                            configuration data is stored. Set to None
                            to load from the default location based
                            on nodeID. Must be seekable, with read/
                            write permission, honor sync requests,
                            and not be written by any other program.
                            Can alternatively be an object implementing
                            the necessary functions to be a crypto
                            config store (load_section, load_value,
                            store_value, load_opaque_value, store_opaque_value,
                            set_cryptokey)
           cryptokey (obj): Optional object implementing the
                            necessary public/private key functions
                            (get/sign_spk,get/use_epk,
                            wrap/unwrap_opaque).
                            Set to None to load from the default
                            location in the configuration file.
        provisioners (obj): Optional object implementing the
                            necessary allowed provisioner functions
                            (reencrypt_request). Set to None to
                            load from the default location in
                            the configuration file.
  """

  def __init__(self, nodeID, kp, kc, config=None, cryptokey=None, provisioners=None):
    super().__init__(nodeID, kp, kc, config, cryptokey)
    if (not ('enable_auto_commit' in self._kc.config) or self._kc.config['enable_auto_commit'] != False):
      self._logger.warning("Auto commit not disabled, controller may miss messages.")
    if (self._kc.config['group_id'] is None):
      self._logger.warning("Group ID not set, controller may miss messages.")
    if (provisioners is None):
      # attempt legacy provisioners load
      provs = self._cryptostore.load_section('provisioners',defaults=False)
      if provs!=None:
        for p in provs:
          k = pysodium.crypto_hash_sha256(provs[p])
          self._cryptostore.store_value(k,provs[p],section='allowlist')
          self._cryptostore.store_value(p,None,section='provisioners')
      allowlist = self._cryptostore.load_section('allowlist',defaults=False)
      if not (allowlist is None):
        allowlist = allowlist.values()
      denylist = self._cryptostore.load_section('denylist',defaults=False)
      if not (denylist is None):
        denylist = denylist.values()
      provisioners = Provisioners(allowlist=allowlist, denylist=denylist)
    if (not hasattr(provisioners, 'reencrypt_request') or not inspect.isroutine(provisioners.reencrypt_request)):
      raise KafkaCryptoControllerError("Invalid provisioners source supplied!")

    self._provisioners = provisioners
    self._last_subscribed_time = 0
    self._mgmt_thread = Thread(target=self._process_mgmt_messages,daemon=True)
    self._mgmt_thread.start()

  # Main background processing loop. Must assume that it can "die" at any
  # time, even mid-stride, so ordering of operations to ensure atomicity and
  # durability is critical.
  def _process_mgmt_messages(self):
    while True:
      # First, (Re)subscribe if needed
      if ((time()-self._last_subscribed_time) >= self.MGMT_SUBSCRIBE_INTERVAL):
        self._logger.debug("Initiating resubscribe...")
        trx = "(.*\\" + self.TOPIC_SUFFIX_SUBS.decode('utf-8') + "$)"
        self._kc.subscribe(topics=[self.MGMT_TOPIC_CHAINS.decode('utf-8'),self.MGMT_TOPIC_ALLOWLIST.decode('utf-8'),self.MGMT_TOPIC_DENYLIST.decode('utf-8')],pattern=trx)
        self._last_subscribed_time = time()
        self._logger.info("Resubscribed to topics.")

      # Second, process messages
      # we are the only thread ever using _kc, _kp, so we do not need the lock to use them
      self._logger.debug("Initiating poll...")
      msgs = self._kc.poll(timeout_ms=self.MGMT_POLL_INTERVAL, max_records=self.MGMT_POLL_RECORDS)
      self._logger.debug("Poll complete with %i msgsets.", len(msgs))
      # but to actually process messages, we need the lock
      for tp,msgset in msgs.items():
        self._logger.debug("Topic %s Partition %i has %i messages", tp.topic, tp.partition, len(msgset))
        self._lock.acquire()
        for msg in msgset:
          self._logger.debug("Processing message: %s", msg)
          topic = msg.topic
          if (isinstance(topic,(str,))):
            topic = topic.encode('utf-8')
          if topic[-len(self.TOPIC_SUFFIX_SUBS):] == self.TOPIC_SUFFIX_SUBS:
            root = topic[:-len(self.TOPIC_SUFFIX_SUBS)]
       	    self._logger.debug("Processing subscribe message, root=%s, msgkey=%s", root, msg.key)
            # New consumer encryption key. Validate
            k,v = self._provisioners.reencrypt_request(root, cryptoexchange=self._cryptoexchange, msgkey=msg.key, msgval=msg.value)
            # Valid request, resign and retransmit
            if (not (k is None)) or (not (v is None)):
              self._logger.info("Valid consumer key request on topic=%s, root=%s, msgkey=%s. Resending to topic=%s, msgkey=%s", topic, root, msg.key, root + self.TOPIC_SUFFIX_REQS, k)
              self._kp.send((root + self.TOPIC_SUFFIX_REQS).decode('utf-8'), key=k, value=v)
            else:
              self._logger.info("Invalid consumer key request on topic=%s, root=%s in message: ", topic, root, msg)
          elif topic == self.MGMT_TOPIC_CHAINS:
            # New candidate public key chain
            self._logger.info("Received new chain message: %s", msg)
            if msg.key == self._cryptokey.get_spk():
              self._logger.debug("Key matches ours. Validating Chain.")
              newchain = self._cryptoexchange.replace_spk_chain(msg.value)
              if not (newchain is None):
                self._logger.info("New chain is superior, using it.")
                self._cryptostore.store_value('chain',newchain,section='crypto')
          elif topic == self.MGMT_TOPIC_ALLOWLIST:
            self._logger.info("Received new allowlist message: %s", msg)
            allow = self._cryptoexchange.add_allowlist(msg.value)
            if not (allow is None):
              c = pysodium.crypto_hash_sha256(allow)
              self._cryptostore.store_value(c,allow,section='allowlist')
          elif topic == self.MGMT_TOPIC_DENYLIST:
            self._logger.info("Received new denylist message: %s", msg)
            deny = self._cryptoexchange.add_denylist(msg.value)
            if not (deny is None):
              c = pysodium.crypto_hash_sha256(deny)
              self._cryptostore.store_value(c,deny,section='denylist')
          else:
            # unknown object
            self._logger.warning("Unknown topic type in message: %s", msg)
        self._lock.release()

      # Third, flush producer
      self._kp.flush()

      # Fourth, commit offsets
      if (self._kc.config['group_id'] is not None):
        self._kc.commit()
  
      # Finally, loop back to poll again
  # end of __process_mgmt_messages




from threading import Lock
import pysodium
import logging
import msgpack
from os import path

class CryptoKey(object):
  """Class utilizing file-backed storage to store (long-term) private key
     material for identity, and memory-backed storage to store (short-term)
     ephemeral private key material.

  Keyword Arguments:
          file (str,file): Filename or File IO object for storing crypto info.
                           Must be readable once.
  """
  #
  # Per instance, defined in init
  #      __file: File object
  #       __spk: signing public key
  #       __ssk: signing private (secret) key
  #        __ek: private key for (un)wrapping opaque bytes
  # Generated ephemerially, on demand:
  #       __epk: dict of encrypting public keys (by topic and usage)
  #       __esk: dict of encrypting private (secret) keys (by topic and usage)
  #
  def __init__(self, file):
    self._logger = logging.getLogger(__name__)
    if (isinstance(file, (str))):
      if (not path.exists(file)):
        self.__init_cryptokey(file)
      with open(file, 'rb') as file:
        data = file.read()
    else:
      data = file.read()
    datalen = len(data)
    # this should be legacy
    contents = None
    while len(data) and contents is None:
      try:
        contents = msgpack.unpackb(data)
      except msgpack.exceptions.ExtraData:
        data = data[:-1]
    if len(data) != datalen:
      self._logger.warning("Cryptokey file had extraneous bytes at end, attempting load anyways.")
    self.__eklock = Lock()
    self.__esk = {}
    self.__epk = {}
    self.__ssk = contents[0]
    self.__spk = pysodium.crypto_sign_sk_to_pk(self.__ssk)
    self.__ek = contents[1]

  def get_spk(self):
    return self.__spk

  def sign_spk(self, msg):
    return pysodium.crypto_sign(msg, self.__ssk)

  def get_epk(self, topic, usage):
    #
    # returns the public key of a new ephemeral encryption key for the specified topic
    #
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    if (isinstance(usage,(str))):
      usage = bytes(usage, 'utf-8')
    with self.__eklock:
      self.__generate_esk(topic, usage)
      return self.__epk[topic][usage]

  def use_epk(self, topic, usage, pks, clear=True):
    rv = []
    if (isinstance(topic,(str))):
      topic = bytes(topic, 'utf-8')
    if (isinstance(usage,(str))):
      usage = bytes(usage, 'utf-8')
    with self.__eklock:
      if not topic in self.__esk or not usage in self.__esk[topic]:
        return rv
      for pk in pks:
        rv.append(pysodium.crypto_scalarmult_curve25519(self.__esk[topic][usage],pk))
      if clear:
        self.__remove_esk(topic, usage)
    return rv

  def wrap_opaque(self, crypto_opaque):
    nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
    return nonce+pysodium.crypto_secretbox(crypto_opaque,nonce,self.__ek)

  def unwrap_opaque(self, crypto_opaque):
    try:
      return pysodium.crypto_secretbox_open(crypto_opaque[pysodium.crypto_secretbox_NONCEBYTES:],crypto_opaque[0:pysodium.crypto_secretbox_NONCEBYTES], self.__ek)
    except:
      return None

  def __generate_esk(self, topic, usage):
    # ephemeral keys are use once only, so always ok to overwrite
    # caller must hold self.__eklock prior to calling
    if not topic in self.__esk or not topic in self.__epk:
      self.__esk[topic] = {}
      self.__epk[topic] = {}
    self.__esk[topic][usage] = pysodium.randombytes(pysodium.crypto_scalarmult_curve25519_BYTES)
    self.__epk[topic][usage] = pysodium.crypto_scalarmult_curve25519_base(self.__esk[topic][usage])

  def __remove_esk(self, topic, usage):
    # caller must hold self.__eklock prior to calling
    self.__esk[topic].pop(usage)
    self.__epk[topic].pop(usage)

  def __init_cryptokey(self, file):
    self._logger.warning("Initializing new CryptoKey file %s", file)
    pk,sk = pysodium.crypto_sign_keypair()
    self._logger.warning("  Public key: %s", pysodium.crypto_sign_sk_to_pk(sk).hex())
    with open(file, "wb") as f:
      f.write(msgpack.packb([sk,pysodium.randombytes(pysodium.crypto_secretbox_KEYBYTES)]))
    self._logger.warning("  CryptoKey Initialized. Provisioning required for successful operation.")



from time import time
import pysodium
import msgpack
import logging
import re

def key_in_list(wanted, choices):
  # choices is an array of zero or more msgpacked arrays, each of
  # which has the format of a chain entry (see below). This helper
  # function iterates each choice, seeing if the key wanted
  # is present, and returns the match (if any)
  # Right now this has O(n) running time in the length of choices,
  # a future improvement is to make the average running time
  # much faster by building a set of the pk's before calling
  # this function
  if wanted is None or choices is None:
    return None
  for c in choices:
    pk = msgpack.unpackb(c)
    if wanted == pk[2]:
      return c
  return None

def multimatch(wanted, choices):
  # choices is an array of zero or more items, and we return
  # True if wanted is matched by one or more items in choices, and
  # False otherwise. Choices is treated as literal
  # strings except if it starts with "^", in which case it is
  # treated as a regular expression. This matches behaviors
  # of many other libraries, including librdkafka.
  # Wanted is also treated as a literal except if it
  # starts with "^", in which case it is treated as a regular
  # expression. wanted is in choices if:
  #  1. Wanted is a literal, and exactly matches a literal in
  #     choices. [OR]
  #  2. Wanted is a literal, and is matched by a regex in
  #     choices. [OR]
  #  3. Wanted is a regex, and exactly matches a regex in
  #     choices.
  # In particular, this means that regexs in choices are only
  # considered to match a regex wanted if the two (treated as
  # literals) exactly match. This is to ensure that wanted
  # is a strict subset of what choices allows. In principle
  # the regexes could be converted to NFA/DFAs and then checked
  # to ensure a wanted regex is a strict subset of a choices
  # regex, but this is presently unimplemented. It does mean
  # that multimatch is, in some cases, more restrictive than
  # strictly necessary, but this is better than being less
  # restrictive (which is a security hazard).
  if wanted is None:
    return True
  if choices is None:
    return False
  for c in choices:
    if len(wanted)<1 or (wanted[0] != b'^'[0] and wanted[0] != '^'):
      # wanted is a literal
      if len(c) > 0 and (c[0] == b'^'[0] or c[0] == '^'):
        r = re.compile(c)
        if r.match(wanted)!=None:
          return True
      elif wanted == c:
        return True
    else:
      # wanted is a regex
      if wanted == c:
        return True
  # went through all possibilities, did not find it
  return False

def intersect_multimatch(set1, set2):
  if set1==None:
    return set2
  if set2==None:
    return set1
  rv = []
  for w in set2:
    if multimatch(w,set1):
      rv.append(w)
  return rv

def validate_poison(name, value, pk):
  # validate that poison allows the given value for the specified
  # condition (name). It is valid if:
  #   1. Name or value is None. [OR]
  #   2. Name is in poison, not on the exception list, 
  #      and value multimatches that condition. [OR]
  #   3. Name is on the exception list, and values matches
  #      as specified for that exception.
  # Otherwise, the condition is invalid and this returns False.
  # Exceptions:
  #   1. Name = "pathlen": valid if pathlen not in poison array,
  #      or if in poison array and value<=that in poison array.
  #
  if name is None or value is None:
    return True
  poison = msgpack.unpackb(pk[1])
  for ku in poison:
    if ku[0] == name:
      if name != b'pathlen' and multimatch(value,ku[1]):
        return True
      elif name == b'pathlen' and value<=ku[1]:
        return True
      else:
        return False
  if name == b'pathlen':
    return True
  return False

def intersect_certs(c1, c2, same_pk):
  # main logic for determining the minimum allowed values consistent
  # with c1 and c2's constraints. Returns an array with the same
  # structure as a chain entry, with 0,1 set as the intersection of
  # c1 and c2, and values 2:...end set from c2
  # To avoid footguns, we make sure to enforce the following
  # additional conditions:
  # 1. If an item in the chain has key-allowlist or key-denylist, it is
  #    also assumed to imply pathlen = 0, unless explicitly overridden.
  pk = []
  if (c2[0] > 0 and (c2[0]<c1[0] or c1[0] == 0)):
    pk.append(c2[0])
  else:
    pk.append(c1[0])
  poisons = [[-1,msgpack.unpackb(c1[1])], [0,msgpack.unpackb(c2[1])]]
  if same_pk:
    poisons = [[0,msgpack.unpackb(c1[1])], [0,msgpack.unpackb(c2[1])]]
  poison_topics = None
  poison_usages = None
  pathlen = None
  for p in poisons:
    pathlen_offset = p[0]
    poison = p[1]
    new_pathlen = None
    for ku in poison:
      if (ku[0] == b'topics'):
        poison_topics = intersect_multimatch(poison_topics, ku[1])
      elif (ku[0] == b'usages'):
        if (multimatch(b'key-allowlist',ku[1]) or multimatch(b'key-denylist',ku[1])) and new_pathlen is None:
          # enforce zero pathlength unless overridden in this poison array
          new_pathlen = 0
        poison_usages = intersect_multimatch(poison_usages, ku[1])
      elif (ku[0] == b'pathlen'):
        new_pathlen = ku[1]
      else:
        raise ValueError("Unknown chain signature poison")
    if new_pathlen!=None and (pathlen is None or (new_pathlen+pathlen_offset)<pathlen):
      pathlen = new_pathlen+pathlen_offset
  # Build new poison array
  poison = []
  if poison_topics!=None:
    poison.append([b'topics',poison_topics])
  if poison_usages!=None:
    poison.append([b'usages',poison_usages])
  if pathlen!=None:
    poison.append([b'pathlen',pathlen])
  pk.append(msgpack.packb(poison))
  # Copy entries 2-infinity
  pk = pk + c2[2:]
  return pk

def process_chain(chain, topic=None, usage=None, allowlist=None, denylist=None):
  #
  # chain should be a single msgpack array, with the next item in the array
  # signed by the entity with public key given in the current item. The first
  # item must be signed by a root of trust (which are given in allowlist). 
  # Each (signed) item is itself a msgpack array containing:
  # (0) a max_age timestamp (64-bit unsigned unix maximum time signature is valid)
  # (1) a poison array. Sets allowed signature key usages, topics, pathlengths.
  # (2) public key
  # Possibly more items in (3)+, especially common for the final item,
  # but not needed to check the chain.
  #
  # Returns a ValueError if chain does not match topic or usage criteria, or
  # uses a denylisted key, unless there is ultimately a chain between
  # an allowlisted key and the final item with no denylisted keys.
  # The final item might itself be the allowlisted key.
  #
  # Otherwise, returns an array containing:
  # (0) The minimum max_age timestamp
  # (1) The intersection of allowed topics and usages, in the
  #     poison array (i.e. the net allowed topics and usages)
  # (2) The public key
  # And any items in (3)+ from the final item.
  #
  if allowlist is None:
    raise ValueError("No possible roots of trust!")
  last_error = ValueError("No roots of trust found!")
  for rot in allowlist:
    try:
      denylisted = False
      val = [rot] + msgpack.unpackb(chain)
      if len(val) < 2:
        # must be validating at least one chain item beyond the item from the allowlist
        raise ValueError("Chain too short!")
      pk = None
      for npk in val:
        if not (pk is None):
          if key_in_list(pk[2],denylist)!=None:
            denylisted = True
            logging.warning("Chain uses denylisted signing key %s.",pk[2].hex())
            # this is not immediately fatal because a subkey might be allowlisted
          elif key_in_list(pk[2],allowlist)!=None:
            # allowlisted subkey overrides denylist
            denylisted = False
            pk = intersect_certs(pk,msgpack.unpackb(key_in_list(pk[2],allowlist)),True)
          try:
            pk = intersect_certs(pk,msgpack.unpackb(pysodium.crypto_sign_open(npk,pk[2])),False)
          except:
            raise ValueError("Invalid signing!")
        else:
          pk = msgpack.unpackb(npk) # root is unsigned
      # must finally check if leaf key is in allowlist, which overrides any denylist entries
      if key_in_list(pk[2],allowlist)!=None:
        denylisted = False
      # make sure our chain doesn't have breaking denylisted keys
      if denylisted:
        raise ValueError("Chain uses denylisted public key")
      # ensure composite chain is valid
      if (pk[0]<time() and pk[0]!=0):
        raise ValueError("Expired Chain!")
      if not validate_poison(b'topics',topic,pk):
        raise ValueError("No matching topic in allowed intersection set.")
      if not validate_poison(b'usages',usage,pk):
        raise ValueError("No matching usage in allowed intersection set.")
      if not validate_poison(b'pathlen',0,pk):
        raise ValueError("Exceeded allowed pathlen.")
      return pk
    except ValueError as e:
      last_error = e
  # We allow a key to self-sign itself to be denylisted. This enables any private
  # key to denylist itself
  try:
    chain = msgpack.unpackb(chain)
    if len(chain) == 2:
      pk = msgpack.unpackb(chain[0])
      pk0 = pk[2]
      pk = intersect_certs(pk,msgpack.unpackb(pysodium.crypto_sign_open(chain[1],pk[2])))
      if pk0 == pk[2] and multimatch(usage, [b'key-denylist']) and validate_poison(b'usages',b'key-denylist',pk) and validate_poison(b'pathlen',0,pk):
        return pk
  except:
    pass
  raise last_error



import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="kafkacrypto",
  version="0.9.9.3",
  license="GNU GPLv2",
  keywords="kafka kafka-crypto kafka-security security crypo",
  author="Tyrel M. McQueen",
  author_email="tmcqueen-pypi@demoivre.com",
  description="Message layer security/crypto for Kafka",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/tmcqueen-materials/kafkacrypto",
  packages=setuptools.find_packages(),
  python_requires='>=3.3',
  install_requires=['pysodium>=0.7.1','msgpack>=0.6.0','kafka-python>=1.4.4'],
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    "Operating System :: OS Independent",
  ],
)
